#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

const uint16_t G[8][16] = {{1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0},
                           {0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1},
                           {0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1},
                           {0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0},
                           {0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 0},
                           {0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1},
                           {0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1},
                           {0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1}};

const uint16_t H[8][16] = {{0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0},
                           {1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0},
                           {1, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0},
                           {1, 1, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0},
                           {0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0},
                           {1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0},
                           {1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0},
                           {0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1}};

int debug = 0;

/**
 * @brief Set the nth bit object
 *
 * @param n index of the bit to set
 * @param m word to modify
 * @return uint16_t
 */
uint16_t set_nth_bit(int n, uint16_t m) {
  short mask = 1 << (15 - n);
  return (m & ~mask) | ((1 << (15 - n)) & mask);
}

// on enleve si ça nous sert jamais a la fin
/**
 * @brief Set the nth bit custom object
 *
 * @param n index of the bit to set
 * @param val value to set
 * @param m word to modify
 * @return uint16_t
 */
uint16_t set_nth_bit_custom(int n, int val, uint16_t m) {
  short mask = 1 << (15 - n);
  return (m & ~mask) | ((val << (15 - n)) & mask);
}

/**
 * @brief Read the nth bit
 *
 * @param n index of the bit to read
 * @param m word to read
 * @return uint16_t
 */
uint16_t get_nth_bit(int n, uint16_t m) { return (m >> (15 - n)) & 1; }

/**
 * @brief Change the nth bit object
 *
 * @param n index of the bit to change
 * @param m word to modify
 * @return uint16_t
 */
uint16_t chg_nth_bit(int n, uint16_t m) { return m ^ (1 << (15 - n)); }

/**
 * @brief Count the number of bits set to 1
 *
 * @param m word to count
 * @return uint8_t
 */
uint8_t cnt_bits(uint16_t m) {
  uint8_t counter = 0;
  for (int i = 0; i <= 15; i++) {
    if (get_nth_bit(i, m) == 1)
      counter++;
  }

  return counter;
}

/**
 * @brief Print a word in binary
 *
 * @param k number of bits to print
 * @param m word to print
 */
void print_word(int k, uint16_t m) {
  fprintf(stdout, "0b");
  for (int i = 15; i >= 15 - k; i--)
    printf("%d", (m >> i) & 1);

  fprintf(stdout, "\n");
}

/**
 * @brief Encode a message with the G matrix
 *
 * @param m message to encode
 * @return uint16_t
 */
uint16_t encode_G(uint16_t m) {
  for (int i = 8; i <= 15; i++) {
    for (int j = 0; j <= 7; j++) {
      if (get_nth_bit(j, m) && G[j][i])
        m = chg_nth_bit(i, m);
    }
  }

  return m;
}

/**
 * @brief calculate the distance correction
 *
 * @return uint8_t
 */
uint8_t dist_code_G() {
  uint16_t dist = 16, counter;

  for (int i = 0; i < 8; i++) {
    for (int j = i + 1; j < 8; j++) {
      counter = 0;

      for (int k = 0; k < 16; k++)
        if (G[i][k] != G[j][k])
          counter++;

      if (counter < dist)
        dist = counter;
    }
  }

  return dist;
}

/**
 * @brief Calculate the correction word
 *
 * @param m word to calculate
 * @return uint8_t
 */
uint8_t mod_poly(uint16_t m) {
  uint16_t dividend = m;
  uint16_t divider = 0b1101110010000000;

  int maxSteps = 8;

  while (maxSteps > 0) {
    // on calcule la division (c'est-à-dire un xor)
    dividend ^= divider;

    while (get_nth_bit(0, dividend) == 0) {
      dividend <<= 1;
      maxSteps--;

      if (maxSteps == 0)
        break;
    }
  }
  return dividend >> 8;
}

/**
 * @brief Encode a message with the polynomial
 *
 * @param m message to encode
 * @return uint16_t message encoded
 */
uint16_t encode_poly(uint16_t m) {
  uint16_t poly = mod_poly(m);
  for (size_t i = 8; i < 15; i++) {
    if (get_nth_bit(i, poly))
      m = chg_nth_bit(i, m);
  }

  return m;
}

/**
 * @brief Check if a word is correct
 *
 * @param m message to decode
 * @return uint8_t 1 if correct 0 if not
 */
uint8_t decode_poly(uint16_t m) {
  if (debug) {
    fprintf(stdout, "_______________________\n");
    fprintf(stdout, "word: \n");
    print_word(15, m);
  }

  // extract the correction code from the word
  uint16_t corCodeWord = m << 8;
  corCodeWord = corCodeWord >> 8;
  if (debug) {
    fprintf(stdout, "corCodeWord: \n");
    print_word(15, corCodeWord);
  }

  // extract the expected correction code from the word
  uint16_t corCodeExpect = m;
  // set the correction code to 0 for encoding
  for (size_t i = 0; i < 8; i++) {
    corCodeExpect = set_nth_bit_custom(15 - i, 0, corCodeExpect);
  }
  corCodeExpect = encode_poly(corCodeExpect) << 8;
  corCodeExpect = corCodeExpect >> 8;
  if (debug) {
    fprintf(stdout, "corCodeExpect: \n");
    print_word(15, corCodeExpect);
  }

  return corCodeWord == corCodeExpect;
}

/**
 * @brief Function correcting a corrupted word (polynomial)
 *
 * @param m
 * @return uint8_t word corrected
 */
uint8_t correct(uint16_t m) {
  if (debug)
    fprintf(stdout, "\n-------Debug correct-------\n");

  uint8_t syndrome = mod_poly(m);

  uint16_t cpy = m;

  for (size_t i = 0; i <= 15; i++) {
    // locate the error
    uint16_t search = mod_poly(0b1000000000000000 >> i);
    if (search == syndrome) {
      cpy = chg_nth_bit(i, cpy);
      if (debug) {
        fprintf(stdout, "cpy : ");
        print_word(15, cpy);
      }
      break;
    }
  }

  return cpy >> 8;
}

/**
 * @brief Function correcting a corrupted word (with matrix H)
 *
 * @param m
 * @return uint8_t word corrected
 */
uint8_t correct_matrix(uint16_t m) {
  if (debug)
    fprintf(stdout, "\n-------Debug correct-------\n");

  uint16_t syndrome = 0b0000000000000000;

  // multiply H*m' to get syndrom
  for (int i = 0; i <= 7; i++) {
    for (int j = 0; j <= 15; j++) {
      if (H[i][j] && get_nth_bit(j, m))
        syndrome = chg_nth_bit(i, syndrome);
    }
    if (debug) {
      fprintf(stdout, "mod L%d : ", i);
      print_word(15, syndrome);
    }
  }

  uint16_t cpy = m;

  for (size_t i = 0; i <= 15; i++) {
    // locate the error
    uint16_t search = mod_poly(0b1000000000000000 >> i);
    if (search == syndrome) {
      cpy = chg_nth_bit(i, cpy);
      if (debug) {
        fprintf(stdout, "cpy : ");
        print_word(15, cpy);
      }
      break;
    }
  }

  return cpy >> 8;
}

/**
 * @brief Function correcting a corrupted word with 2 errors as specified in the
 * subject
 *
 * @param m
 * @return unsigned char
 */
unsigned char correct2errors(uint16_t m) {
  if (debug)
    fprintf(stdout, "\n-------Debug correct2errors-------\n");

  uint8_t syndrome = mod_poly(m);

  if (debug) {
    fprintf(stdout, "syndrome : ");
    print_word(7, syndrome << 8);
  }

  uint16_t cpy = m;

  if (syndrome == 0b00000000) {
    return m >> 8;
  } else {
    int twoError = 1;
    for (size_t i = 0; i <= 15; i++) {
      // locate the error
      uint8_t search = mod_poly(0b1000000000000000 >> i);
      if (search == syndrome) {
        cpy = chg_nth_bit(i, cpy);
        if (debug) {
          fprintf(stdout, "cpy : ");
          print_word(15, cpy);
        }
        break;
      }
    }

    if (twoError) {
      for (size_t i = 0; i <= 15; i++) {
        for (size_t j = 0; j <= 15; j++) {
          {
            // locate the error
            uint8_t search =
                mod_poly((0b1000000000000000 >> i) ^ (0b1000000000000000 >> j));
            if (search == syndrome) {

              if (rand() % 3 == 0) {
                cpy = chg_nth_bit(i, cpy);
              } else if (rand() % 3 == 1) {
                cpy = chg_nth_bit(j, cpy);
              } else {
                cpy = chg_nth_bit(i, cpy) | chg_nth_bit(j, cpy);
              }

              if (debug) {
                fprintf(stdout, "cpy : ");
                print_word(15, cpy);
              }
              break;
            }
          }
        }
      }
    }
  }
  return cpy >> 8;
}

/**
 * @brief Function correcting a corrupted word with 2 errors as specified in the
 * subject
 *
 * @param m
 * @return unsigned char
 */
unsigned char correct2errorsTxt(uint16_t m) {
  if (debug)
    fprintf(stdout, "\n-------Debug correct2errors-------\n");

  uint8_t syndrome = mod_poly(m);

  if (debug) {
    fprintf(stdout, "syndrome : ");
    print_word(7, syndrome << 8);
  }

  uint16_t cpy = m;

  if (syndrome == 0b00000000) {
    return m >> 8;
  } else {
    int twoError = 1;
    for (size_t i = 0; i <= 15; i++) {
      // locate the error
      uint8_t search = mod_poly(0b1000000000000000 >> i);
      if (search == syndrome) {
        cpy = chg_nth_bit(i, cpy);
        if (debug) {
          fprintf(stdout, "cpy : ");
          print_word(15, cpy);
        }
        break;
      }
    }

    if (twoError) {
      for (size_t i = 0; i <= 15; i++) {
        for (size_t j = 0; j <= 15; j++) {
          {
            // locate the error
            uint8_t search =
                mod_poly((0b1000000000000000 >> i) ^ (0b1000000000000000 >> j));
            if (search == syndrome) {
              uint16_t test = chg_nth_bit(i, cpy) | chg_nth_bit(j, cpy);
              // check if the bit is a letter
              if (97 <= (test >> 8) && (test >> 8) <= 122) {
                cpy = test;
                // check if the bit is a special char
              } else if ((test >> 8) == 27 || (test >> 8) == 32 ||
                         (test >> 8) == 44 || (test >> 8) == 46) {
                cpy = test;
              } else {
                test = chg_nth_bit(i, cpy);
                if (97 <= (test >> 8) && (test >> 8) <= 122) {
                  cpy = test;
                } else if ((test >> 8) == 27 || (test >> 8) == 32 ||
                           (test >> 8) == 44 || (test >> 8) == 46) {
                  cpy = test;
                } else {
                  test = chg_nth_bit(j, cpy);
                  if (97 <= (test >> 8) && (test >> 8) <= 122) {
                    cpy = test;
                  } else if ((test >> 8) == 27 || (test >> 8) == 32 ||
                             (test >> 8) == 44 || (test >> 8) == 46) {
                    cpy = test;
                  } else {
                    if (rand() % 3 == 0) {
                      cpy = chg_nth_bit(i, cpy);
                    } else if (rand() % 3 == 1) {
                      cpy = chg_nth_bit(j, cpy);
                    } else {
                      cpy = chg_nth_bit(i, cpy) | chg_nth_bit(j, cpy);
                    }
                  }
                }
              }

              if (debug) {
                fprintf(stdout, "cpy : ");
                print_word(15, cpy);
              }
              break;
            }
          }
        }
      }
    }
  }
  return cpy >> 8;
}

/**
 * @brief Main function with tests
 *
 * @param arcg
 * @param argv
 * @return int
 */
int main(int argc, char **argv) {
  if (argc == 2 && argv[1][0])
    debug = 1;

  srand(time(NULL));

  uint16_t word = 0b1101010011001011;
  fprintf(stdout, "-----test print word -----\n");
  fprintf(stdout, "word: ");
  print_word(15, word);

  fprintf(stdout, "-----test chg_nth_bit-----\n");
  fprintf(stdout, "word before: ");
  print_word(15, word);
  word = chg_nth_bit(5, word);
  fprintf(stdout, "target 6bit: ");
  print_word(15, word);

  fprintf(stdout, "-----test set_nth_bit-----\n");
  fprintf(stdout, "word before: ");
  print_word(15, word);
  fprintf(stdout, "target 3bit: ");
  word = set_nth_bit(2, word);
  print_word(15, word);

  fprintf(stdout, "-----test get_nth_bit-----\n");
  fprintf(stdout, "word: ");
  print_word(15, word);
  fprintf(stdout, "test: 0b");
  for (size_t i = 0; i <= 15; i++) {
    fprintf(stdout, "%d", get_nth_bit(i, word));
  }
  fprintf(stdout, "\n");

  fprintf(stdout, "-----test count cnt_bits-----\n");
  fprintf(stdout, "expected : 8\n");
  fprintf(stdout, "returned : %d\n", cnt_bits(0b1111000011110000));

  fprintf(stdout, "-----test dist_code_G-----\n");
  fprintf(stdout, "expected : 4\n");
  fprintf(stdout, "returned : %d\n", dist_code_G());

  /*
  4.3) Etant donne que la distance de Hamming est de 4,
  on peut detecter jusqu'a 3 erreurs avec ce code.
  Cependant, on ne peut corriger qu'une seule erreur,
  car pour 2 erreurs, on ne sait pas (sans probabilites)
  quel etait le mot d'origine. Ce qui montre qu'une
  distance de Hamming paire n'est pas tres pertinente.
  */

  fprintf(stdout, "-----test encode_G-----\n");
  fprintf(stdout, "original word: ");
  uint16_t word1 = 0b0011100100000000;
  print_word(15, word1);
  fprintf(stdout, "encoded  word: ");
  print_word(15, encode_G(word1));
  fprintf(stdout, "expected word: 0b0011100110011010\n");

  fprintf(stdout, "-----test mod_poly-----\n");
  fprintf(stdout, "expected : 0b00101100\n");
  fprintf(stdout, "returned : ");
  uint16_t word2 = 0b0000101100000000;
  print_word(7, mod_poly(word2)
                    << 8); // bit shift because the print is from the left
  // Q 7
  fprintf(stdout, "-----test encode_poly-----\n");
  fprintf(stdout, "   raw  word : ");
  print_word(15, word1);
  fprintf(stdout, "encoded word : ");
  print_word(15, encode_poly(word1));

  fprintf(stdout, "-----test decode_poly-----\n");
  fprintf(stdout, "expected 1 : %d | word: ", decode_poly(encode_poly(word2)));
  print_word(15, encode_poly(word2));
  fprintf(stdout, "expected 0 : %d | word: ",
          decode_poly(chg_nth_bit(0, encode_poly(word2))));
  print_word(15, chg_nth_bit(0, encode_poly(word2)));
  fprintf(stdout, "expected 0 : %d | word: ",
          decode_poly(chg_nth_bit(15, encode_poly(word2))));
  print_word(15, chg_nth_bit(15, encode_poly(word2)));
  // Q8
  fprintf(stdout, "-----test Q8-----\n");
  fprintf(stdout, "test word : ");
  print_word(15, chg_nth_bit(0, encode_poly(word2)));
  fprintf(stdout, "mod poly  : ");
  print_word(7, mod_poly(chg_nth_bit(0, encode_poly(word2))) << 8);
  fprintf(stdout, "test word : ");
  print_word(15, chg_nth_bit(0, encode_poly(word1)));
  fprintf(stdout, "mod poly  : ");
  print_word(7, mod_poly(chg_nth_bit(0, encode_poly(word1))) << 8);
  fprintf(stdout, "test word : 0b1000000000000000\n");
  fprintf(stdout, "mod poly  : ");
  print_word(7, encode_poly(0b1000000000000000) << 8);

  /* Q9 ?
  On peut faire un décodage par syndrome avec un code polynomial car le passage
  du mot dans mod poly renvois le syndrome et donc la position de l'erreur. La
  raison est que le premier passage dans mod poly encode le mot et le deuxième
  envoie le syndrome.
  */
  fprintf(stdout, "-----Justification Q9-----\n");
  fprintf(stdout, "test word   : ");
  print_word(15, word2);
  fprintf(stdout, "encoded poly: ");
  uint16_t encoded = encode_poly(word2);
  print_word(15, encoded);
  fprintf(stdout, "syndrome    : ");
  print_word(7, mod_poly(encoded) << 8);
  fprintf(stdout, "wth err at 0: ");
  print_word(7, mod_poly(chg_nth_bit(0, encoded)) << 8);

  // Q 10
  fprintf(stdout, "-----test correct-----\n");
  uint16_t word3 = 0b1011100110011010;
  fprintf(stdout, "expected word : 0b00111001\n");
  fprintf(stdout, "   correct    : ");
  print_word(7, correct(word3) << 8);
  /*
  Pour correct(), tout d'abord j'utilise mod_poly pour obtenir le syndrome.
  Ensuite, je compare le syndrome avec les syndromes possibles (définis dans le
  tableau syndromes) pour déterminé la position de l'erreur, et enfin je corrige
  grace a la position determine. Pour correct_matrix(), je fais la meme chose
  mais en obtenant le syndrome via la multiplication H*m'. J'ai choisit de
  prendre la methode par polinome car elle est plus simple a implementer et
  engendre moin de calcules.
  */

  // Q 11
  /*
  1. Non, car avec une distance de Hamming de 4 (déterminé dans la question 4),
  on peut donc détecter 3 erreurs et corriger jusqu'a 1 erreur car la 2ème
  erreur est ambigue.
  2. Non, on peut pas faire de décodage par syndrome dans tous les cas car il
  faut que la distance de Hamming soit impaire. Oui, il existe peut être des
  situations ou la deuxième erreur est corrigeable correctement.
  3. Oui, le code est parfait car le mieu qu'on puisse corriger éfficacement sur
  une détection de 3 erreurs est 1 erreur.
  */

  // Q 12
  fprintf(stdout, "-----test correct2errors-----\n");
  fprintf(stdout, "expected word : 0b00111001\n");
  // add an error
  uint16_t word4 =
      chg_nth_bit(5, word3) & chg_nth_bit(7, word3) & chg_nth_bit(0, word3);
  fprintf(stdout, "  2 erreurs   : ");
  print_word(7, correct2errors(word4) << 8);
  fprintf(stdout, "  1 erreurs   : ");
  print_word(7, correct2errors(word3) << 8);
  fprintf(stdout, "  sans erreur : ");
  print_word(7, correct2errors(chg_nth_bit(0, word3)) << 8);

  // Q14

  // Q15
  FILE *fd;
  fd = fopen("./assets/corrupted_file_RL", "r");
  if (fd == NULL) {
    fprintf(stderr, "Error opening file\n");
    exit(1);
  }

  fprintf(stdout, "-----test corrupted_file_RL-----\n");
  fprintf(stdout, "original file: \n");
  uint16_t data;
  size_t last = fread(&data, sizeof(uint16_t), 1, fd);
  do {
    fprintf(stdout, "%c", data);
  } while (last == fread(&data, sizeof(uint16_t), 1, fd));

  fprintf(stdout, "\ncorrected file: \n");
  fd = fopen("./assets/corrupted_file_RL", "r");
  last = fread(&data, sizeof(uint16_t), 1, fd);
  do {
    fprintf(stdout, "%c", correct2errors(data));
  } while (last == fread(&data, sizeof(uint16_t), 1, fd));

  fprintf(stdout, "\ntext optimised: \n");
  fd = fopen("./assets/corrupted_file_RL", "r");
  last = fread(&data, sizeof(uint16_t), 1, fd);
  do {
    fprintf(stdout, "%c", correct2errorsTxt(data));
  } while (last == fread(&data, sizeof(uint16_t), 1, fd));

  fprintf(stdout, "\n-----BONUS----- \n");

  fclose(fd);

  /*----------------------BONUS----------------------------*/

  // open the meme
  FILE *fd2;
  fd2 = fopen("./assets/9.jpg", "r");
  if (fd2 == NULL) {
    fprintf(stderr, "Error opening file 2\n");
    exit(1);
  }

  // create the output file
  FILE *fd3;
  fd3 = fopen("./output/meme_corrige.jpg", "w");
  if (fd3 == NULL) {
    fprintf(stderr, "Error opening file 3\n");
    exit(1);
  }

  // add errors to the data
  unsigned char data2;
  size_t last2;
  do {
    last2 = fread(&data2, sizeof(unsigned char), 1, fd2);

    uint16_t buff = encode_poly(data2 << 8);
    // add errors
    if (rand() % 100 > 98) {
      buff = chg_nth_bit(rand() % 8, buff) | chg_nth_bit(rand() % 8, buff);
    }
    fprintf(fd3, "%c", correct2errors(buff));
  } while (last2 > 0);

  fclose(fd2);
  fclose(fd3);

  return 0;
}
